// --------------------------------------------------------------------------------------------------------------------
//  Copyright Unai Landa, part of MicroLib(uLib)
// 
// --------------------------------------------------------------------------------------------------------------------
#include "gp32.h"
#include "..\uColor.h"
#include "..\uSprite.h"

// opcodes
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
template<typename EXE_ONCE,typename EXE_TWICE>
struct  PIXEL_OP
{
  static  inline  void  Make1   (u16 &uDst, const u16 &uSrc, u32 uTran, u32 uParm)  { uDst = EXE_ONCE::Exe(uDst,uSrc,uParm);    }
  static  inline  void  Make11  (u32 &uDst, const u32 &uSrc, u32 uTran, u32 uParm)  { uDst = EXE_TWICE::Exe(uDst,uSrc,uParm);   }
};
template<typename EXE_ONCE,typename EXE_TWICE>
struct  PIXEL_OPT
{
  static  inline  void  Make1   (u16 &uDst, const u16 &uSrc, u32 uTran, u32 uParm)  { if(uSrc != uTran)  uDst = EXE_ONCE::Exe(uDst,uSrc,uParm);  }
  static  inline  void  Make11  (u32 &uDst, const u32 &uSrc, u32 uTran, u32 uParm)  { if(uSrc != uTran)
                                                                                      {
                                                                                        if( ((u16)uSrc) == ((u16)uTran) )
                                                                                          uDst = (uDst&0x0000ffff) | (EXE_ONCE::Exe(uDst>>16,uSrc>>16,uParm)<<16);
                                                                                        else
                                                                                        if( (uSrc>>16) == ((u16)uTran) )
                                                                                          uDst = (uDst&0xffff0000) | (EXE_ONCE::Exe(uDst,uSrc,uParm)&0xffff);
                                                                                        else
                                                                                          uDst = EXE_TWICE::Exe(uDst,uSrc,uParm);
                                                                                      }
                                                                                    }
};

//  PixelOp
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
typedef PIXEL_OP <ColorSpace5551::SetOnce,ColorSpace5551::SetTwice>     OP_Set;
typedef PIXEL_OPT<ColorSpace5551::SetOnce,ColorSpace5551::SetTwice>     OP_SetT;

typedef PIXEL_OP <ColorSpace5551::AddOnce,ColorSpace5551::AddTwice>     OP_Add;
typedef PIXEL_OPT<ColorSpace5551::AddOnce,ColorSpace5551::AddTwice>     OP_AddT;

typedef PIXEL_OP <ColorSpace5551::SubOnce,ColorSpace5551::SubTwice>     OP_Sub;
typedef PIXEL_OPT<ColorSpace5551::SubOnce,ColorSpace5551::SubTwice>     OP_SubT;

typedef PIXEL_OP <ColorSpace5551::BlendOnce,ColorSpace5551::BlendTwice> OP_Blend;
typedef PIXEL_OPT<ColorSpace5551::BlendOnce,ColorSpace5551::BlendTwice> OP_BlendT;
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
static  inline  void  _ClipSprite(u16* &pSrc, u16* &pDst, int iX, int iY, int &iSizeX, int &iSizeY, u32 &uSrc, u32 &uDst)
{
  //  clip Y
  int yb = 0; int ye = iSizeY;
  if(iY<0)
    yb -= iY, iY = 0;
  if(iY+(ye-yb) >= 320)
    ye = 320-(iY+yb);

  //  clip X
  int xb = 0; int xe = iSizeX;
  if(iX<0)
    xb -= iX, iX = 0;
  if(iX+(xe-xb) >= 240)
    xe = 240-(iX+xb);

  pDst += iY*240    + iX;
  pSrc += yb*iSizeX + xb;

  //  salto
  uSrc = iSizeX;
  uDst = 240;
  
  //  tamao
  iSizeX = xe-xb;
  iSizeY = ye-yb;
}



// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
template<typename OP> inline
void _DrawSpriteADAO (u16 *pSprite, u16 *pBuffer, int iX, int iY, int iSizeX, int iSizeY, u32 uSrc, u32 uDst, u32 uTran, u32 uAlpha)
{
  bool bI = iSizeX&1;
  iSizeX&=~1;
  u16*  pSrc = (pSprite+(iSizeX));
  u16*  pDst = (pBuffer+(iSizeX));

  iSizeX  =-iSizeX>>1;

  uTran &= 0xffff;
  uTran |= uTran<<16;

  do
  {
    int xx = iSizeX;
    do
    {
      OP::Make11( ((u32*)pDst)[xx], ((u32*)pSrc)[xx], uTran, uAlpha);
    }while(++xx);
  
    if(bI)
      OP::Make1( pDst[0], pSrc[0], uTran, uAlpha);
      
    pDst += uDst;
    pSrc += uSrc;
  }while(--iSizeY);
  
}

// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
template<typename OP> inline
void _DrawSpriteUDUO  (u16 *pSprite, u16 *pBuffer, int iX, int iY, int iSizeX, int iSizeY, u32 uSrc, u32 uDst, u32 uTran, u32 uAlpha)
{
  u16*  pSrc = (pSprite+iSizeX);
  u16*  pDst = (pBuffer+iSizeX);

  iSizeX=-iSizeX;
  do
  {
    int xx = iSizeX;
    do
    {
      OP::Make1(pDst[xx],pSrc[xx],uTran,uAlpha);
    }while(++xx);
    
    pDst += uDst;
    pSrc += uSrc;
  }while(--iSizeY);

}

template<typename OP> inline
void _DrawSprite  (u16 *pSprite, u16 *pBuffer, int iX, int iY, int iSizeX, int iSizeY, u32 uTran, u32 uAlpha)
{
  u32   uSrc,uDst;
  _ClipSprite(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst);

  if(iSizeX<=0 || iSizeY<=0)
    return;

  if( !(((u32)pBuffer)&3) && !(((u32)pSprite)&3) )
    _DrawSpriteADAO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
  else
  if( !(((u32)pBuffer)&3) )
  //_DrawSpriteADUO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
    _DrawSpriteUDUO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
  else
  if( !(((u32)pSprite)&3) )
  //_DrawSpriteUDAO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
    _DrawSpriteUDUO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
  else
    _DrawSpriteUDUO<OP>(pSprite, pBuffer, iX, iY, iSizeX, iSizeY, uSrc, uDst, uTran, uAlpha);
}

// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
void uS_DrawSprite    (u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_Set>(sprite, framebuffer, put_x, put_y, xsize, ysize, 0, 0);
}


void uS_DrawSpriteA   ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_Add>(sprite, framebuffer, put_x, put_y, xsize, ysize, 0, 0);
}


void uS_DrawSpriteS   ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_Sub>(sprite, framebuffer, put_x, put_y, xsize, ysize, 0, 0);
}


void uS_DrawSpriteB   ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 xsize, u16 ysize, u8 alpha)
{
  _DrawSprite<OP_Blend>(sprite, framebuffer, put_x, put_y, xsize, ysize, 0, alpha);
}


void uS_DrawSpriteT   ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_SetT>(sprite, framebuffer, put_x, put_y, xsize, ysize, trans, 0);
}


void uS_DrawSpriteTA  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_AddT>(sprite, framebuffer, put_x, put_y, xsize, ysize, trans, 0);
}


void uS_DrawSpriteTS  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans, u16 xsize, u16 ysize)
{
  _DrawSprite<OP_SubT>(sprite, framebuffer, put_x, put_y, xsize, ysize, trans, 0);
}


void uS_DrawSpriteTB  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans, u16 xsize, u16 ysize, u8 alpha)
{
  _DrawSprite<OP_BlendT>(sprite, framebuffer, put_x, put_y, xsize, ysize, trans, alpha);
}


//  compatibilidad MrMirko
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
typedef struct {
   char magic[4];
   u16 size_x;
   u16 size_y;
   u16 info;
   u16 alpha;
   //u16 raw[0];
} SHEADER;
/*
info:
format :[0-3]   0000 RGB  16 Format RRRRRGGGGGBBBBB-   normal sprite data
                0001 RGB  16 Format RRRRRGGGGGBBBBB-   normal sprite data + blend in data
                0010 ARGB 32 Format -----------AAAAARRRRRGGGGGBBBBB-
                0011 ...

reserved:[4-5]    reserved, always 0
rotate:6          0 Sprite is rotated  ( 90 deg clockwise )
                  1 Sprite is not rotated
alpha:7           0 Alpha/Blend disable
                  1 Alpha/Blend enabled
*/

// --------------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------------
void uS_DrawSpriteH   ( u16 *sprite, short put_x, short put_y, u16 *framebuffer)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_Set>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, 0, 0);
}

void uS_DrawSpriteHA  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_Add>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, 0, 0);
}

void uS_DrawSpriteHS  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_Sub>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, 0, 0);
}

void uS_DrawSpriteHB ( u16 *sprite, short put_x, short put_y, u16 *framebuffer,u8 alpha)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_Blend>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, 0, alpha);
}

void uS_DrawSpriteHT  ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_SetT>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, trans, 0);
}

void uS_DrawSpriteHTA ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_AddT>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, trans, 0);
}

void uS_DrawSpriteHTS ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_SubT>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, trans, 0);
}

void uS_DrawSpriteHTB ( u16 *sprite, short put_x, short put_y, u16 *framebuffer, u16 trans, u8 alpha)
{
  SHEADER *pHdr = (SHEADER *)sprite;
  _DrawSprite<OP_BlendT>(sprite+6, framebuffer, put_x, put_y, pHdr->size_x, pHdr->size_y, trans, alpha);
}

